/* 
 *                      ... The Multicomputer Toolbox ...
 *
 * test_prog/csparsetest/csparsetest1.c                 *** Host Program ***
 * Csparse sparse matrix library tester component.
 * by Anthony Skjellum and Alvin P. Leung
 *
 * 
 * Anthony Skjellum
 * Lawrence Livermore National Laboratory
 * Numerical Mathematics Group
 * 7000 East Avenue, L-316
 * Livermore, CA 94550
 * (415)422-1161
 * FAX: (415)423-2993
 * tony@helios.llnl.gov
 *
 * .   .   .   .   .   .   .   .   .
 * 
 * 
 * (C) 1989, 1990, 1991 California Institute of Technology, Inc.
 * (C) 1991 Regents of the University of California.
 * All rights reserved.
 * 
 * Zipcode version history
 * 
 * created:     **-Sep-89 {APL}
 * last update: 02-Apr-91 {AS} ... 29-Mar-90 based on ... 27-Feb-90 {AS}
 * 
 * 02-Apr-91:   provide support for IPSC_860 model variation {AS}
 *              use better zipcode cohort generation; R flag disabled
 *              for now but still valid ... nodal crash detection is useless.
 *
 * 01-Apr-90:   nodal crash detection added.
 *
 * 29-Mar-90:   removed spurious letter free from the main() routine...
 *              no "Letter" variable actually needed there... deleted it
 *              and variable Letter_Size cosequently from main()'s list.
 *
 * 24-Mar-90:   if there is an extra node, spawn Zipcode.pmg.node by itself
 *              on node 0 (ie, start with node 1 for real work).  This is
 *              strictly to avoid wasting memory and let more examples run
 *              correctly.  Because we are running on a heterogeneous
 *              machine (some nodes have 3mb, some have 4mb), and because
 *              our advanced data distributions favor placing the load
 *              imbalance in low-numbered grid processes, and because high
 *              numbered physical nodes tend to have 4mb of memory, do 
 *              a simple reversal on the ordering in the addressee list.
 *              On a homogeneous memory machine, this reversal shouldn't
 *              significantly alter performance.
 *
 * 21-Mar-90:   support freeing of nodes at termination through a new query
 * 11-Mar-90:   add support for more data distributions
 * 04-Mar-90:   add option to use no datafile and concurrently generate
 *              the matrix (also updates in csparsetest1a.c, correspondingly)
 * 03-Mar-90:   add missing gettimeofday() calls for setting "tim" struct
 * 02-Mar-90:   support direct reading of compressed files via pipe
 * 28-Feb-90:   continue...
 * 27-Feb-90:   basically, rewrite this entire code for the new sparse matrix
 *              library, and improve its behavior insofar as timing accuracy.
 *
 *              First, re-do the code so that the host transmits the sparse
 *              matrix directly, instead of a dense representation of the 
 *              sparse matrix.
 */


#include <stdio.h>
/*
#include <ctype.h>
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>

#include <csparse.h>
#include <cstats.h>
#include <csparsetest.h>


#define DEBUG
#define GETCUBE   0

#ifndef IPSC860
#define COHORT_FIRST_NODE 0
#else
#define COHORT_FIRST_NODE 1
#endif
#define FALSE 0
#define TRUE ~FALSE

#if 0
#define DEBUG0    1
#define DEBUG1    1
#define DEBUG2    1
#define DEBUG3    1
#define DEBUG4    1
#define DEBUG5    1
#endif

#define STASIS

#define FILENAME_LEN 256
#define COMMAND_LEN   80
#define PROGNAME_LEN 100

/* Bugs get around for cc */

MAIN_() {}
/* Bugs get around for cc */

main(argc, argv)
int argc;
char *argv[];
{
    /* getopt() support: */
    STASIS int c;
    extern int optind;
    extern char *optarg;

    /* ... */
    STASIS int i;

    STASIS ZIP_ADDRESSEES *Addr_List;
    STASIS ZIP_MAILER     *Mailer EQ_NULL;

    static char            Command[COMMAND_LEN], *Answer;
    static char            DataFile_Name[FILENAME_LEN];
    static char            DataFile_NameZ[FILENAME_LEN];
    static char            StatFile_Name[FILENAME_LEN],
                           stat_file_name[FILENAME_LEN];

    static char            deflt_mach0[COMMAND_LEN];

    STASIS int             file_format_type;

    STASIS int             Counter, Node, Index EQ_ZERO;
    STASIS int             P EQ_ZERO;
    STASIS int             Q EQ_ZERO;
    STASIS int             p, q;
    STASIS int             PQ EQ_ZERO;
    STASIS int             number_of_nodes EQ_ZERO;
    STASIS long            M EQ_ZERO;
    STASIS long            N EQ_ZERO;

#if 0 /* used? */
    STASIS int             Local_Max_Row EQ_ZERO;
    STASIS int             Local_Max_Col EQ_ZERO;
    STASIS int             Local_Row_Index EQ_ZERO;
    STASIS int             Local_Col_Index EQ_ZERO;
#endif

    STASIS Cdistrib       *Row_Dis EQ_NULL;
    STASIS void          (*Row_Distribution)() EQ_NULL;
    STASIS int           (*Row_Inverse_Distribution)() EQ_NULL;
    STASIS int           (*Row_Distribution_limit)() EQ_NULL;
    STASIS int           (*Row_Distribution_init)() EQ_NULL;
    STASIS void           *Row_Extra = NULL;
    STASIS long            Row_Extra_param1 EQ_ZERO, Row_Extra_param2 EQ_ZERO;
    STASIS char           *Row_Extra_param1_name EQ_NULL,
                          *Row_Extra_param2_name EQ_NULL;

    STASIS Cdistrib       *Col_Dis EQ_NULL;
    STASIS void          (*Col_Distribution)() EQ_NULL;
    STASIS int           (*Col_Inverse_Distribution)() EQ_NULL;
    STASIS int           (*Col_Distribution_limit)() EQ_NULL;
    STASIS int           (*Col_Distribution_init)() EQ_NULL;
    STASIS void           *Col_Extra = NULL;
    STASIS long            Col_Extra_param1 EQ_ZERO, Col_Extra_param2 EQ_ZERO;
    STASIS char           *Col_Extra_param1_name EQ_NULL,
                          *Col_Extra_param2_name EQ_NULL;

#if 0 /* used? */
    STASIS double         *Local_RHS EQ_NULL;
#endif

    STASIS short           Row_Dis_Type EQ_ZERO;
    STASIS short           Col_Dis_Type EQ_ZERO;
    STASIS short          *Dis_Header EQ_NULL;
    STASIS long           *Mat_Data_Header EQ_NULL;

#if 0 /* used? */
    STASIS int             Global_Row_Index EQ_ZERO;
    STASIS int             Global_Col_Index EQ_ZERO;
    STASIS int             Matrix_Row EQ_ZERO;
    STASIS int             Matrix_Col EQ_ZERO;
#endif

    STASIS int             Size_Row_extra EQ_ZERO;
    STASIS int             Size_Col_extra EQ_ZERO;

    STASIS long            Repeat EQ_ZERO;
    STASIS long            Strategy EQ_ZERO;
    STASIS long            Dump_Freq EQ_ZERO, Regen_Flag EQ_ZERO;

    STASIS long            Upper_Band_Width, Lower_Band_Width, skew;
    STASIS double          Fill_Percent, scale, sp_scale;
    
    STASIS long            Matrix_Entries EQ_ZERO, RHS_Entries EQ_ZERO;

    STASIS FILE           *DataFile EQ_NULL;
    STASIS FILE           *StatFile EQ_NULL;

    STASIS int             len;
    STASIS char           *ptr EQ_NULL;

    /* statistics: */
    STASIS Cstats *mean_stats = NULL;
    STASIS STATS  *max_stats = NULL, *min_stats = NULL;

    /* date/time tagging feature: */
#ifdef GETTIMEOFDAY
    STASIS struct timeval tim;
#endif
    STASIS char *str_ptr EQ_NULL; /* used to help in getting time string */

    /* file status info for extracting data of ages of the executables... */
    STASIS char *host_prog_age EQ_NULL, *node_prog_age EQ_NULL;
    struct stat *stat_buf EQ_NULL;

    /* ... */

    static char node_pgm_name[PROGNAME_LEN], aux_node_pgm_name[PROGNAME_LEN];
    
    /* ... */

    STASIS long dump_count, nstats;
    /* ... */
    STASIS int contnu_flag = 0;
    STASIS int context = 0;

    STASIS int pre_seed = 0;

    /* ... */

    STASIS int freenodes_flag = 0;
    STASIS int killnodes_flag = 0;
    STASIS int crash_result = 0;
    /* ... */

    STASIS int Zflag = 0, Rflag = 0, Nnodes = 0, Cflag = 0, Kflag = 1,
                          Sflag = 0;

    /* ... */

#ifdef IPSC860
      char *deflt_mach = "ipsc860";
#else
#ifdef GHOST
      char *deflt_mach = "gh";
#else
      char *deflt_mach = "s2010";
#endif
#endif

    /* ... */

    /* immediately print any output to stdout: */
    setbuf(stdout,0);

    /* ... */

    /* cosmic_init(ZIP_HNODE, ZIP_HPID); */

    /* ... */

    strcpy(node_pgm_name, "csparsetest1a");      /* default */
    stat_file_name[0] = '\0';

    while ((c = getopt(argc, argv, "CMRKZM:p:S:")) != EOF)
    {
        switch (c)
        {
#if 0
            case 'C':   Cflag = 1; break;       /* crash testing */
#endif
            case 'C':   break;

            case 'K':   Kflag = 0; break;       /* active low */
            case 'R':   Rflag = 1; break;
            case 'Z':   Zflag = 1; break;
            case 'M':   strncpy(deflt_mach0, optarg, COMMAND_LEN);
                        deflt_mach = deflt_mach0;
                        break;
            case 'p':   strncpy(node_pgm_name, optarg, COMMAND_LEN); 
                        break;
            case 'S':   strncpy(stat_file_name, optarg, COMMAND_LEN); 
                        Sflag = 1;
                        break;
            case '?':   printf("usage: %s [-M machine] [-p node_prog] [-S statfile] [-K] [-Z]\n");
                        exit(1);
        }
    }

    /* ... */
    sprintf(aux_node_pgm_name, "%s.%s",node_pgm_name, deflt_mach);

    /* get info. on age of this program for statistical reference */
#ifdef GETTIMEOFDAY
    stat_buf = NEW_CLEAR_ITEM(struct stat,1);
    stat(argv[0], stat_buf);
    str_ptr = asctime(localtime(&(stat_buf -> st_mtime)));
    str_ptr[strlen(str_ptr)-1] = '\0'; /* strip newline */
#else
    str_ptr = "(unknown)";
#endif
    host_prog_age = NEW_CLEAR_ITEM(char, 1+strlen(str_ptr));
    strcpy(host_prog_age, str_ptr);

#ifdef GETTIMEOFDAY
    stat(aux_node_pgm_name, stat_buf);
    str_ptr = asctime(localtime(&(stat_buf -> st_mtime)));
    str_ptr[strlen(str_ptr)-1] = '\0'; /* strip newline */
#else
    str_ptr = "(unknown)";
#endif
    node_prog_age = NEW_CLEAR_ITEM(char, 1+strlen(str_ptr));
    strcpy(node_prog_age, str_ptr);

    /* ... */

    /* get time of day for tagging: */
#ifdef GETTIMEOFDAY
    gettimeofday(&tim, NULL);
#endif

    printf("Csparsetest1.c -- main tester for Csparse\n");
 
    printf("Logical 2D Grid dimensions (eg, 4 8): ");
    scanf("%d %d", &P, &Q);
    printf("%d %d\n", P, Q);

    number_of_nodes = PQ = P * Q;

#ifndef IPSC860
#ifdef GETCUBE
#ifdef IPSC860
   ++number_of_nodes;
#endif

    sprintf(Command, "getcube %s %dn", deflt_mach, number_of_nodes);
    if (system(Command))
        if(Kflag)
           ckill(-1, -1, "");
#else
    if(Kflag)
        ckill(-1, -1, "");
#endif
#endif

#ifdef IPSC860
    if((PQ < (Nnodes-1)) && (Zflag))
#else
    if((PQ < Nnodes) && (Zflag))
#endif
        Zflag = 1;
    else
        Zflag = 0;

    if (Zip_init() == NULL)
    {
       printf("csparsetest1: error - zip_init\n");
       exit(1);
    }
   
    Nnodes = zip_nnodes();
#ifdef DEBUG
    printf("csparsetest1: Nnodes = %d vs PQ = %d\n", Nnodes, PQ);
#endif

    Addr_List = zip_new_cohort(PQ, COHORT_FIRST_NODE, COHORT_PID, TRUE);

    /* spawn cohort of processes: */
    Zip_spawn(node_pgm_name, Addr_List);

    /* mailer for cohort: */
    if ((Mailer = g2_grid_open(&P, &Q, Addr_List)) == NULL)
    {
        printf("csparsetest1: error - failed to open grid mailer\n");
        exit(1);
    }

#if DEBUG1
    print("csparsetest1 -- start");
#endif

    Zip_push(Mailer);

#if DEBUG1
    print("csparsetest1 -- Checkpoint 1");
#endif

    /* set "pre-seed" for generating seeds for random generators: */

    printf("Enter random number pre-seed or -1 to use a time-based pre-seed ");
    scanf("%d", &pre_seed);
    if(pre_seed == -1)
       pre_seed = time(NULL);
    printf("%d\n", pre_seed);
    SRAND(pre_seed);

#if DEBUG1
    printf("%d,%d: csparsetest1 -- past SRAND()\n", mynode(), mypid());
#endif

    for(;;)
    {
        printf("what is the name for the statistics (append-output) file? ");
        scanf("%s", StatFile_Name);
        if(Sflag)                               /* override it */
            strcpy(StatFile_Name, stat_file_name);
        printf("%s\n",StatFile_Name);

        /* append on end of file to avoid unwanted loss of stat. data: */
        if ((StatFile = fopen(StatFile_Name, "a")) == NULL)
            printf("Cannot open data file %s\n", StatFile_Name);
        else
        {
#ifndef IPSC860
            setlinebuf(StatFile);       /* see each line as written */
#endif
            break;
        }
    }

    printf("Pivoting strategy:\n");
    printf(" %5d - no pivoting\n",CMATRIX_NO_PIVOTING);
    printf(" %5d - partial row pivoting\n",CMATRIX_PARTIAL_ROW_PIVOTING);
    printf(" %5d - partial column pivoting\n",CMATRIX_PARTIAL_COLUMN_PIVOTING);
    printf(" %5d - diagonal pivoting\n",CMATRIX_PARTIAL_DIAGONAL_PIVOTING);
    printf(" %5d - complete pivoting\n",CMATRIX_COMPLETE_PIVOTING);
    printf(" %5d - preset pivoting\n",CMATRIX_PRESET_PIVOTING);
    printf(" %5d - multi row pivoting\n",CMATRIX_MULTI_ROW_PIVOTING);
    printf(" %5d - multi column pivoting\n",CMATRIX_MULTI_COLUMN_PIVOTING);
    printf(" %5d - markowitz count pivoting [not supported presently]\n",
                                CMATRIX_MARKOWITZ_COUNT_PIVOTING);

    printf("Pivoting strategy -> ");
    scanf("%ld", &Strategy);
    printf("%ld\n",Strategy);

    printf("Matrix row and column data distributions:\n");
    printf(" %5d - Linear\n",CDISTRI_LINEAR);
    printf(" %5d - Scattered\n",CDISTRI_SCATTERED);
    printf(" %5d - Block Linear  [div. constraint]\n",CDISTRI_BLINEAR);
    printf(" %5d - Block Scatter [div. constraint]\n",CDISTRI_BSCATTERED);
    printf(" %5d - Genl. Block Linear\n", CDISTRI_GBLINEAR);
    printf(" %5d - Genl. Block Scatter\n", CDISTRI_GBSCATTERED);
    printf(" %5d - One-Parameter Distribution\n", CDISTRI_G1PARAM);
    printf(" %5d - Two-Parameter Distribution\n", CDISTRI_G2PARAM);
    printf("Row Dis., Col. Dis -> ");
    scanf("%hd %hd", &Row_Dis_Type, &Col_Dis_Type);
    printf("%d %d\n", Row_Dis_Type, Col_Dis_Type);

    /* defaults */
    Row_Extra_param1_name = "*";
    Row_Extra_param2_name = "*";
    Row_Extra_param1 = 0;
    Row_Extra_param2 = 0;

    if ((Row_Dis_Type == CDISTRI_BLINEAR)     ||
        (Row_Dis_Type == CDISTRI_BSCATTERED)  ||
        (Row_Dis_Type == CDISTRI_GBLINEAR)    ||
        (Row_Dis_Type == CDISTRI_GBSCATTERED) ||
        (Row_Dis_Type == CDISTRI_G2PARAM))
    {
        printf("Row data distribution Block (B) factor = ");
        scanf("%ld", &Row_Extra_param1);
        printf("%ld\n", Row_Extra_param1);

        Row_Extra_param1_name = "B";
    }

    if ((Row_Dis_Type == CDISTRI_G1PARAM) || (Row_Dis_Type == CDISTRI_G2PARAM))
    {
        printf("Row data distribution Scattering (S) Parameter = ");
        scanf("%ld", &Row_Extra_param2);
        printf("%ld\n", Row_Extra_param2);

        Row_Extra_param2_name = "S";
    }

    /* create row distribution: */
    Row_Dis = new_Cdistrib(Mailer, CDISTRI_ROW, Row_Dis_Type, NULL);
    Row_Distribution         = Row_Dis -> mu;
    Row_Inverse_Distribution = Row_Dis -> mu_i;
    Row_Distribution_limit   = Row_Dis -> mu_lim;
    Row_Distribution_init    = Row_Dis -> mu_init;

    /* initialize the distribution accordingly: */
    switch(Row_Dis_Type)
    {
        case CDISTRI_BLINEAR:
        case CDISTRI_BSCATTERED:
        case CDISTRI_GBLINEAR:
        case CDISTRI_GBSCATTERED:
                Row_Extra = (void *)NEW_CLEAR_ITEM(int ,1);
                *(int *)Row_Extra = Row_Extra_param1;
                break;
        case CDISTRI_G1PARAM:
                Row_Extra = (void *)NEW_CLEAR_ITEM(
                                CDISTRI_GENL_1PARAM_EXTRA, 1);
                Row_Distribution_init(Row_Extra,
                        Row_Extra_param2, P, M);
                break;
        case CDISTRI_G2PARAM:
                Row_Extra = (void *)NEW_CLEAR_ITEM(
                                CDISTRI_GENL_2PARAM_EXTRA, 1);
                Row_Distribution_init(Row_Extra, Row_Extra_param1,
                        Row_Extra_param2, P, M);
                break;
        default:
                break;
    }
    if(Row_Dis -> mu_extra != NULL)
    {
        Free(Row_Dis -> mu_extra);
        Row_Dis -> mu_extra = Row_Extra;
    }

    /* defaults */
    Col_Extra_param1_name = "*";
    Col_Extra_param2_name = "*";
    Col_Extra_param1 = 0;
    Col_Extra_param2 = 0;

    if ((Col_Dis_Type == CDISTRI_BLINEAR)     ||
        (Col_Dis_Type == CDISTRI_BSCATTERED)  ||
        (Col_Dis_Type == CDISTRI_GBLINEAR)    ||
        (Col_Dis_Type == CDISTRI_GBSCATTERED) ||
        (Col_Dis_Type == CDISTRI_G2PARAM))
    {
        printf("Column data distribution Block (B) factor = ");
        scanf("%ld", &Col_Extra_param1);
        printf("%ld\n", Col_Extra_param1);  

        Col_Extra_param1_name = "B";
    }

    if ((Col_Dis_Type == CDISTRI_G1PARAM) || (Col_Dis_Type == CDISTRI_G2PARAM))
    {
        printf("Col data distribution Scattering (S) Parameter = ");
        scanf("%ld", &Col_Extra_param2);
        printf("%ld\n", Col_Extra_param2);

        Col_Extra_param2_name = "S";
    }

    /* create column distribution: */
    Col_Dis = new_Cdistrib(Mailer, CDISTRI_COLUMN, Col_Dis_Type, NULL);
    Col_Distribution         = Col_Dis -> mu;
    Col_Inverse_Distribution = Col_Dis -> mu_i;
    Col_Distribution_limit   = Col_Dis -> mu_lim;
    Col_Distribution_init    = Col_Dis -> mu_init;

    /* initialize the distribution accordingly: */
    switch(Col_Dis_Type)
    {
        case CDISTRI_BLINEAR:
        case CDISTRI_BSCATTERED:
        case CDISTRI_GBLINEAR:
        case CDISTRI_GBSCATTERED:
                Col_Extra = (void *)NEW_CLEAR_ITEM(int ,1);
                *(int *)Col_Extra = Col_Extra_param1;
                break;
        case CDISTRI_G1PARAM:
                Col_Extra = (void *)NEW_CLEAR_ITEM(
                                CDISTRI_GENL_1PARAM_EXTRA, 1);
                Col_Distribution_init(Col_Extra,
                        Col_Extra_param2, Q, N);
                break;
        case CDISTRI_G2PARAM:
                Col_Extra = (void *)NEW_CLEAR_ITEM(
                                CDISTRI_GENL_2PARAM_EXTRA, 1);
                Col_Distribution_init(Col_Extra, Col_Extra_param1,
                        Col_Extra_param2, Q, N);
                break;
        default:
                break;
    }
    if(Col_Dis -> mu_extra != NULL)
    {
        Free(Col_Dis -> mu_extra);
        Col_Dis -> mu_extra = Col_Extra;
    }

    /* ... */

    printf("Repetitions & Dump Frequency: ");
    scanf("%ld %ld", &Repeat, &Dump_Freq);
    printf("%ld %ld\n", Repeat, Dump_Freq);
    printf("`Random Matrix Regeneration' (y/n) ");
    scanf("%s", Command); Answer = Command;
    while((isspace(*Answer)) && (*Answer != '\0')) ++Answer;

#if DEBUG0
    printf("*Answer = %c\n",*Answer);
#endif

    *Answer = (isupper(*Answer)) ? tolower(*Answer) : *Answer;

#if DEBUG0
    printf("*Answer = %c\n",*Answer);
#endif

    if(*Answer == 'y')
    {
        printf("Yes.\n");
        Regen_Flag = 1;
    }
    else
    {
        printf("No.\n");
        Regen_Flag = 0;
    }

    printf("%ld\n", Regen_Flag);

    /* ... */

#ifdef GETTIMEOFDAY
    str_ptr = asctime(localtime(&(tim . tv_sec)));
    str_ptr[strlen(str_ptr)-1] = '\0'; /* strip newline */
#else
    str_ptr = "(unknown)";
#endif

    print("#Csparsetest1 START @ %s",str_ptr);
    print("#Csparsetest1 HOST CODE %s of %s", argv[0], host_prog_age);
    print("#Csparsetest1 NODE CODE %s of %s", node_pgm_name, node_prog_age);
    print("#Csparsetest1: P = %d; Q = %d; Nnodes = %d; Pre_Seed = %d",
                P, Q, nnodes(), pre_seed);
    print("#Csparsetest1: Row_Dis = %d, Col_Dis = %d",Row_Dis_Type,
                                                        Col_Dis_Type);
    print("#Csparsetest1: Row_Params: %s=%d, %s=%d, Col_Params: %s=%d, %s=%d",
        Row_Extra_param1_name, Row_Extra_param1,
        Row_Extra_param2_name, Row_Extra_param2,
        Col_Extra_param1_name, Col_Extra_param1,
        Col_Extra_param2_name, Col_Extra_param2);

    print("#Csparsetest1: Statfile = [%s]",StatFile_Name);

    /* put comparable information directly in the statistics file: */
    fprintf(StatFile, "%% START RUN @ %s\n", str_ptr);
    fprintf(StatFile, "%% HOST CODE %s of %s\n%% NODE CODE %s of %s\n",
                argv[0], host_prog_age, node_pgm_name, node_prog_age);
    fprintf(StatFile, "%% Name of this run is = [%s]\n",StatFile_Name);
    fprintf(StatFile, "%% P = %d; Q = %d; Nnodes = %d; Pre_Seed = %d\n",
                P, Q, nnodes(), pre_seed);
    fprintf(StatFile, "%% Row_Dis = %d, Col_Dis = %d\n",Row_Dis_Type,
                                                        Col_Dis_Type);
    fprintf(StatFile, "%% Row_Params: %s=%d, %s=%d, Col_Params: %s=%d, %s=%d\n",
        Row_Extra_param1_name, Row_Extra_param1,
        Row_Extra_param2_name, Row_Extra_param2,
        Col_Extra_param1_name, Col_Extra_param1,
        Col_Extra_param2_name, Col_Extra_param2);

#ifdef DEBUG1
    print("csparsetest1 -- before matrix file loop");
#endif

    for(;;)
    {
        printf("Use an%s input file? ",(contnu_flag) ? "other" : "");
        scanf("%s", Command); Answer = Command;
        while((isspace(*Answer)) && (*Answer != '\0')) ++Answer;

#if DEBUG0
        printf("*Answer = %c\n",*Answer);
#endif

        *Answer = (isupper(*Answer)) ? tolower(*Answer) : *Answer;

#if DEBUG0
        printf("*Answer = %c\n",*Answer);
#endif

        if(*Answer == 'y')
            printf("Yes.\n");
        else
        {
            printf("No.\n");
            break;
        }

#define BINNING 1000
        matrix_read_and_dist_interface(Mailer, Row_Dis_Type, Col_Dis_Type, 
            &M, &N, Row_Extra_param1,
                Row_Extra_param2, Col_Extra_param1, Col_Extra_param2,
                Row_Extra, Col_Extra, Repeat, Strategy, Dump_Freq, Regen_Flag,
                contnu_flag, BINNING, context++);
    }

    if(!contnu_flag)
    {
        printf("Enter the Matrix Dimensions (M N) ");
        scanf("%d %d", &M, &N);
        printf("%d %d\n", M, N);

#if DEBUG0
        print("transmitting intro letter...");
#endif
        transmit_intro_letter(Mailer, Row_Dis_Type, Col_Dis_Type,
                        M, N,
                        Row_Extra_param1, Row_Extra_param2, 
                        Col_Extra_param1, Col_Extra_param2, Repeat,
                        Strategy, Dump_Freq, Regen_Flag);

#if DEBUG0
        print("done transmitting intro letter...");
#endif
    }

    /* transmit random-seeds to all grid processes: */
    for(p = 0; p < P; p++)
        for(q = 0; q < Q; q++)
            transmit_seed_letter(Mailer, (int)LRAND(), p, q);

    contnu_flag = 0;
    for(;;)
    {
        printf("Perform an%s sband generation? ",(contnu_flag) ? "other" : "");
        scanf("%s", Command); Answer = Command;
        while((isspace(*Answer)) && (*Answer != '\0')) ++Answer;

#if DEBUG0
        printf("*Answer = %c\n",*Answer);
#endif

        *Answer = (isupper(*Answer)) ? tolower(*Answer) : *Answer;

#if DEBUG0
        printf("*Answer = %c\n",*Answer);
#endif

        if(*Answer == 'y')
            printf("Yes.\n");
        else
        {
            printf("No.\n");
            break;
        }

        printf("Enter Upper Band Width (incl. 'diagonal') ");
        scanf("%ld", &Upper_Band_Width); printf("%d\n", Upper_Band_Width);
        printf("Enter Lower Band Width (excl. 'diagonal') ");
        scanf("%ld", &Lower_Band_Width); printf("%d\n", Lower_Band_Width);
        printf("Enter Skew for band, 0: no skew, + ==> NE, - ==>  SW ");
        scanf("%ld", &skew); printf("%d\n", skew);
        printf("Enter Fill percent for off_band sprinkle  (0.0 ==> none) ");
        scanf("%lf", &Fill_Percent); printf("%g\n", Fill_Percent);
        printf("Enter scale for band and sprinkle entries: ");
        scanf("%lf %lf", &scale, &sp_scale); printf("%g %g\n",scale,sp_scale);

#if 0
        print("#SBAND[%d]: Lower BW = %d, Upper BW = %d, skew = %d",
                context, Lower_Band_Width, Upper_Band_Width, skew);
        print("#SBAND[%d]: scale = %g, Sp. Fill %% = %g, Sp. scale = %g",
                context, scale, Fill_Percent, sp_scale);
#endif
        printf("%d,%d: #SBAND[%d]: Lower BW = %d, Upper BW = %d, skew = %d\n",
                mynode(), mypid(),
                context, Lower_Band_Width, Upper_Band_Width, skew);
printf("%d,%d: #SBAND[%d]: scale = %g, Sp. Fill %% = %g, Sp. scale = %g\n",
                mynode(), mypid(),
                context, scale, Fill_Percent, sp_scale);

        /* comparable output directly to StatFile: */
        fprintf(StatFile,
                "%% SBAND[%d]: Lower BW = %d, Upper BW = %d, skew = %d\n",
                context, Lower_Band_Width, Upper_Band_Width, skew);
        fprintf(StatFile,
                "%% SBAND[%d]: scale = %g, Sp. Fill %% = %g, Sp. scale = %g\n",
                context, scale, Fill_Percent, sp_scale);
        fflush(StatFile);
        
#ifdef DEBUG
printf("%d,%d: calling transmit_sband() sequence...\n",mynode(), mypid());
#endif

        for(p = 0; p < P; p++)
            for(q = 0; q < Q; q++)
                transmit_sband_gen_letter(Mailer, Upper_Band_Width,
                        Lower_Band_Width, skew, Fill_Percent, scale, sp_scale,
                        context, p, q);

#ifdef DEBUG
printf("%d,%d: done calling transmit_sband() sequence...\n",mynode(), mypid());
#endif

        ++context;
        ++contnu_flag;
    }

#ifdef DEBUG1
    print("csparsetest1 -- after all matrix generation(s)");
#endif
    for(;;)
    {
        printf("Free nodes when done? ");
        scanf("%s", Command); Answer = Command;
        while((isspace(*Answer)) && (*Answer != '\0')) ++Answer;

#if DEBUG0
        printf("*Answer = %c\n",*Answer);
#endif

        *Answer = (isupper(*Answer)) ? tolower(*Answer) : *Answer;

#if DEBUG0
        printf("*Answer = %c\n",*Answer);
#endif

        if(*Answer == 'y')
        {
            printf("Yes.\n");
            freenodes_flag = 1;
            break;
        }
        else
        {
            printf("No.\n");
            break;
        }

    }


    /* clue in the nodes that they've got all their data... */
    for(p = 0; p < P; p++)
        for(q = 0; q < Q; q++)
            transmit_end_of_dbase_letter(Mailer, p, q);

#ifdef DEBUG1
    print("csparsetest1 -- after transmit_end_of_dbase_letter's");
#endif

    print("#Csparsetest1: M = %ld; N = %ld; Read Init. Entries [A/rhs] = %ld, %ld", M, N, Matrix_Entries, RHS_Entries);
    print("#Csparsetest1: Repetitions = %ld, Piv. Strategy = %ld, Freq = %ld",
        Repeat, Strategy, Dump_Freq);
    print("#Csparsetest1: Random Regen. flag = %ld", Regen_Flag);

    /* comparable info. to stat file: */
    fprintf(StatFile, "%% M = %ld; N = %ld; Read Init. Entries [A/rhs] = %ld, %ld\n", M, N, Matrix_Entries, RHS_Entries);
    fprintf(StatFile, "%% Repetitions = %ld, Piv. Strategy = %ld, Freq = %ld\n",
        Repeat, Strategy, Dump_Freq);
    fprintf(StatFile, "%% Random Regen. flag = %ld\n\n", Regen_Flag);

    if(Cflag)
        init_test_for_crash();

    for(;;)     /* await return of statistics letters: */
    {
        /* crash testing alarm setup... */
        if(Cflag)
        {
            if(reset_test_for_crash_sem4())
                alarm(900);     /* 15 minutes till next test */
            
            if((crash_result = get_test_for_crash_result()) > 0)
            {
                print("#Csparsetest1 Nodal Crash detected!");
                fprintf(StatFile, "Nodal Crash detected!\n\n");

                if(crash_result == 2)
                {
                    killnodes_flag = 1;
                    freenodes_flag = 0;
                }

                if(crash_result == 1)
                {
                    killnodes_flag = 0;
                    freenodes_flag = 1;
                }

                break;
            }
        }

#if DEBUG1 && 0
        print("trying load_stats...");
#endif

        /* statistics loading... */
        if(!load_stats_from_nodes(Mailer, &dump_count, &nstats, &mean_stats,
                                                &min_stats, &max_stats))
            continue;

#if DEBUG1
        print("past load_stats...calling dump_stats...");
#endif

        /* dump latest statistics to file: */
        dump_stats_to_file(StatFile, dump_count, nstats, mean_stats,
                                                min_stats, max_stats);

#if DEBUG1
        print("past dump_stats...calling dump_stats...");
#endif

        /* dump stats to host stream too ("print" mode): */
        if(dump_count == -1)
            dump_stats_to_output(dump_count, nstats, mean_stats, min_stats,
                                                max_stats);

#if DEBUG1
        print("past dump_stats...through test...");
#endif

        /* are we through? */
        if(dump_count == -1)
            break;
    }

    /* get time of day for tagging: */
#ifdef GETTIMEOFDAY
    gettimeofday(&tim, NULL);
    
    str_ptr = asctime(localtime(&(tim . tv_sec)));
#else
    str_ptr = "(unknown)";
#endif

    print("#Csparsetest1 END @ %s", str_ptr);
    print("#Csparsetest1 Goodbye.");

    /* closing comments to statistics file too: */
    fprintf(StatFile, "\n%% END @ %s\n", str_ptr);
    fprintf(StatFile, "%% Goodbye.\n\n");

    fclose(StatFile);  /* close statistics file */

    Zip_pop();
    zip_exit();

#if DEBUG1
    print("csparsetest1 -- done");
#endif

#ifndef IPSC860
    if(freenodes_flag)
        system("freecube");
    if(killnodes_flag)
        system("killcube");
#endif
}

matrix_read_and_dist_interface(Mailer, Row_Dis_Type, Col_Dis_Type, 
    M, N, 
    Row_Extra_param1, Row_Extra_param2, Col_Extra_param1, Col_Extra_param2,
    Row_Extra, Col_Extra,
    Repeat, Strategy, Dump_Freq, Regen_Flag, contnu_flag, binning, context)

ZIP_MAILER *Mailer;
int Row_Dis_Type, Col_Dis_Type;
int *M, *N;
long Row_Extra_param1, Row_Extra_param2;
long Col_Extra_param1, Col_Extra_param2;
void *Row_Extra, *Col_Extra;
long Repeat, Strategy, Dump_Freq, Regen_Flag;
int *contnu_flag;
int binning, context;
{
    FILE                  *DataFile;
    static char            DataFile_Name[256];
    static char            DataFile_NameZ[256];
    static char            pipe_cmd[256];
    STASIS int             uncompress_file = 0;
    char *ptr;
    int len;
    int error = 0;

    int Matrix_Entries, RHS_Entries;

    for(;;)
    {
        printf("what is the name of the %sdata file (containing A/RHS)? ",
                (*contnu_flag) ? "next " : "");
        scanf("%s", DataFile_Name); 
        len = strlen(DataFile_Name);
        ptr = DataFile_Name + len - 1;

        /* trailing space stripping... */
        while(((*ptr == ' ')||(*ptr == '\t'))&&(ptr > DataFile_Name))
            --ptr;
        if((*ptr == 'Z')&&(ptr[-1] == '.'))
            uncompress_file = 1;

        printf("%s\n",DataFile_Name);

        if (!uncompress_file &&
                (DataFile = fopen(DataFile_Name, "r")) == NULL)
        {
            sprintf(DataFile_NameZ, "%s.Z",DataFile_Name);

            if ((DataFile = fopen(DataFile_NameZ, "r")) == NULL)
            {
                printf("Cannot open compressed or regular data file %s\n",
                    DataFile_Name);

                continue;
            }
            else
            {
                fclose(DataFile);
                uncompress_file = 1;
            }
        }

#ifndef IPSC860
        if(uncompress_file)
        {
            sprintf(pipe_cmd, "zcat %s", DataFile_Name);

            if ((DataFile = popen(pipe_cmd, "r")) == NULL)
            {
                printf("Cannot open compressed file input pipe of %s",
                        DataFile_Name);
            }
            else
                break;
        }
#endif
    }

    print("#Csparsetest1: %dth Datafile read = [%s]",context, DataFile_Name);

    if(!*contnu_flag)
    {
        int file_format_type;
        double rand_scale;

        Read_Size_of_Sparse_Matrix_Data(DataFile, &M,
                        &N, &rand_scale, &file_format_type);
        transmit_intro_letter(Mailer, Row_Dis_Type, Col_Dis_Type,
                        M, N, Row_Extra_param1,
                        Row_Extra_param2, Col_Extra_param1, Col_Extra_param2,
                        Repeat, Strategy, Dump_Freq, Regen_Flag);
    }

    if (Read_and_Distribute_Sparse_Matrix_Data(DataFile, Mailer,
        Row_Dis_Type, Col_Dis_Type, Row_Extra, Col_Extra,
        &Matrix_Entries, &RHS_Entries, binning, context, *contnu_flag))
    {
        printf("Error reading matrix data from file %s",DataFile_Name);
        error = 1;
    }

    if(!uncompress_file)
        fclose(DataFile);
    else
        pclose(DataFile);

    (*contnu_flag)++;

    return(error);
}

#define DEBUG_LOAD_STATS_FROM_NODES 0
int load_stats_from_nodes(Mailer, dump_count, nstats, mean_stats,
                                                    min_stats, max_stats)
ZIP_MAILER *Mailer;
long *dump_count;
long *nstats; 
Cstats **mean_stats;
STATS **min_stats, **max_stats;
{
    int  i;
    char *Letter EQ_NULL;
    long *Header EQ_NULL;
    long Letter_Size;
    long n_longs;
    long Letter_Type;

    short  *Stats_short EQ_NULL;
    long   *Stats_long  EQ_NULL;
    double *Stats_double EQ_NULL;

    int     short_index;
    int     long_index;
    int     double_index;

    int     Return_Value;

    CSTATS *global_stats EQ_NULL;
    STATS  *the_stats EQ_NULL;

#if DEBUG_LOAD_STATS_FROM_NODES
char subroutine_name = "load_stats_from_nodes()";

print("%s entry... Mailer = %lxX", subroutine_name, Mailer);
#endif

    /* statistics available??? */
    /* receive from corner of grid */
    if ((Letter = g2_recvb(Mailer, 0, 0)) == NULL) 
        return(0);

    Zip_push(Mailer);

#if DEBUG_LOAD_STATS_FROM_NODES
print("%s received possible letter...", subroutine_name);
#endif

    Header = (long *)Letter;

    if (Header[0] == STAT_LETTER)
    {

       Letter_Size = Header[1];
       n_longs     = Header[2];
       *nstats     = Header[3];
       *dump_count = (int)Header[4];

#if DEBUG_LOAD_STATS_FROM_NODES
print("%s: Letter_Size = %ld, n_longs = %ld, *nstats = %ld, *dump_count = %ld",
      subroutine_name, Letter_Size, n_longs, *nstats, *dump_count);
#endif

       Stats_short  = (short *)(Header + n_longs);
       Stats_long   = (long *)(Stats_short + *nstats * SHORTS_PER_STAT);
       Stats_double = (double *)(Stats_long + *nstats * LONGS_PER_STAT);

       if (*mean_stats == NULL)
       {
          *mean_stats = NEW_CLEAR_ITEM(Cstats, *nstats);

#if DEBUG_LOAD_STATS_FROM_NODES
print("%s Allocated *mean_stats = %lxX", subroutine_name, *mean_stats);
#endif
       }

       if (*mean_stats != NULL)
       {
          for (i = 0; i < *nstats; i++)
          {
#if DEBUG_LOAD_STATS_FROM_NODES
print("%s: top of unparing mean stats i = %d", subroutine_name, i);
#endif
             /* overall min/max and where achieved: */
             /* global mean distribution: */
             global_stats = &((*mean_stats)[i].global_stats);
             the_stats    = &(global_stats -> stats);

#if DEBUG_LOAD_STATS_FROM_NODES
print("mean: global_stats = %lxX, the_stats = %lxX", global_stats, the_stats);
#endif
             short_index = i * SHORTS_PER_STAT;
             ctohs(&(Stats_short[0]), SHORTS_PER_STAT);

             global_stats -> min_p = Stats_short[short_index];
             global_stats -> min_q = Stats_short[short_index + 1];
             global_stats -> max_p = Stats_short[short_index + 2];
             global_stats -> max_q = Stats_short[short_index + 3];

             long_index = i * LONGS_PER_STAT;

             ctohl(&(Stats_long[long_index]), 1);

             the_stats -> n = Stats_long[long_index];

             double_index = i * REALS_PER_STAT;
             ctohd(&(Stats_double[double_index]), 6);

             the_stats -> min_value   = Stats_double[double_index];
             the_stats -> max_value   = Stats_double[double_index + 1];
             the_stats -> mean        = Stats_double[double_index + 2];
             the_stats -> sigma       = Stats_double[double_index + 3];
             the_stats -> sigma_mean  = Stats_double[double_index + 4];
             the_stats -> sum_squared = Stats_double[double_index + 5];
          }
       }

       if (*min_stats == NULL)
       {
          *min_stats = NEW_CLEAR_ITEM(STATS, *nstats);

#if DEBUG_LOAD_STATS_FROM_NODES
print("%s Allocated *min_stats = %lxX", subroutine_name, *min_stats);
#endif
       }

       if (*min_stats != NULL)
       {
          for (i = 0; i < *nstats; i++)
          {
#if DEBUG_LOAD_STATS_FROM_NODES
print("%s: top of unparing min stats i = %d", subroutine_name, i);
#endif
             the_stats = &((*min_stats)[i]);

#if DEBUG_LOAD_STATS_FROM_NODES
print("min: the_stats = %lxX", the_stats);
#endif
             /* distribution of the min's: */
             long_index = (i * LONGS_PER_STAT) + 1;
             ctohl(&(Stats_long[long_index]), 1);

             the_stats -> n = Stats_long[long_index];

             double_index = (i * REALS_PER_STAT) + 6;
             ctohd(&(Stats_double[double_index]), 4);

             the_stats -> mean        = Stats_double[double_index];
             the_stats -> sigma       = Stats_double[double_index + 1];
             the_stats -> sigma_mean  = Stats_double[double_index + 2];
             the_stats -> sum_squared = Stats_double[double_index + 3];
          }
       }

       if (*max_stats == NULL)
       {
          *max_stats = NEW_CLEAR_ITEM(STATS, *nstats);

#if DEBUG_LOAD_STATS_FROM_NODES
print("%s Allocated *max_stats = %lxX", subroutine_name, *max_stats);
#endif
       }

       if (*max_stats != NULL)
       {
          for (i = 0; i < *nstats; i++)
          {
#if DEBUG_LOAD_STATS_FROM_NODES
print("%s: top of unparing max stats i = %d", subroutine_name, i);
#endif
             the_stats = &((*max_stats)[i]);

#if DEBUG_LOAD_STATS_FROM_NODES
print("max: the_stats = %lxX", the_stats);
#endif
             long_index = (i * LONGS_PER_STAT) + 2;
             ctohl(&(Stats_long[long_index]), 1);

             the_stats -> n = Stats_long[long_index];

             double_index = (i * REALS_PER_STAT) + 10;
             ctohd(&(Stats_double[double_index]), 4);

             the_stats -> mean        = Stats_double[double_index];
             the_stats -> sigma       = Stats_double[double_index + 1];
             the_stats -> sigma_mean  = Stats_double[double_index + 2];
             the_stats -> sum_squared = Stats_double[double_index + 3];
          }
       }

       Return_Value = 1;
    }
    else
    {
       print("#Csparsetest1: received %d type letter, ignoring...",
               Header[0]);
       Return_Value = 0;
    }

#if DEBUG_LOAD_STATS_FROM_NODES
    print("load_stats_from_nodes() exit...");
#endif

    zip_free(Mailer, Letter);
    return(Return_Value);
}


#define DEBUG_DUMP_STATS_TO_FILE 0
dump_stats_to_file(StatFile, dump_count, nstats, mean_stats,
                                                min_stats, max_stats)
FILE *StatFile;
long dump_count, nstats;
Cstats *mean_stats;
STATS *min_stats, *max_stats;
{
    int i;
    static prefix[100];
    CSTATS *global_stats;
    STATS  *the_stats;

#if DEBUG_DUMP_STATS_TO_FILE
    print("dump_stats_to_file(): entry, StatFile = %lxX",StatFile);
#endif

    for(i = 0; i < nstats; i++)
    {
        global_stats = &(mean_stats[i].global_stats);
        the_stats    = &(global_stats -> stats);

        sprintf(prefix,"%5d # %2d #",dump_count, i);    /* make line prefix */
        fprintf(StatFile,"%s Extr: %5ld # %13.6e %3d %3d %13.6e %3d %3d\n",
          prefix, the_stats -> n,the_stats -> min_value,
          global_stats -> min_p, global_stats -> min_q, the_stats -> max_value,
          global_stats -> max_p, global_stats -> max_q);
        fprintf(StatFile,"%s Mean: %5ld # %13.6e %13.6e %13.6e %13.6e\n",
          prefix, the_stats -> n, 
          the_stats -> mean, the_stats -> sigma_mean, the_stats -> sigma,
                the_stats -> sum_squared);
        
        the_stats = min_stats + i;
        fprintf(StatFile,"%s MinD: %5ld # %13.6e %13.6e %13.6e %13.6e\n",
            prefix, the_stats -> n, the_stats -> mean,
            the_stats -> sigma_mean, the_stats -> sigma,
                the_stats -> sum_squared);
                
        the_stats = max_stats + i;
        fprintf(StatFile,"%s MaxD: %5ld # %13.6e %13.6e %13.6e %13.6e\n\n",
            prefix, the_stats -> n, the_stats -> mean,
            the_stats -> sigma_mean, the_stats -> sigma,
                the_stats -> sum_squared);
                
    }

#if DEBUG_DUMP_STATS_TO_FILE
    print("dump_stats_to_file(): exit.");
#endif
}

#define DEBUG_DUMP_STATS_TO_OUTPUT 0
dump_stats_to_output(dump_count, nstats, mean_stats, min_stats, max_stats)
long dump_count, nstats;
Cstats *mean_stats;
STATS *min_stats, *max_stats;
{
    int i;
    static prefix[100];
    CSTATS *global_stats;
    STATS  *the_stats;

#if DEBUG_DUMP_STATS_TO_OUTPUT
    print("dump_stats_to_output(): entry.");
#endif

    for(i = 0; i < nstats; i++)
    {
        global_stats = &(mean_stats[i].global_stats);
        the_stats    = &(global_stats -> stats);

        sprintf(prefix,"%5d # %5d #",dump_count, i);    /* make line prefix */

        print("%s Extr: %5ld # %13.6e %3d %3d %13.6e %3d %3d",
          prefix, the_stats -> n,the_stats -> min_value,
          global_stats -> min_p, global_stats -> min_q, the_stats -> max_value,
          global_stats -> max_p, global_stats -> max_q);

        print("%s Mean: %5ld # %13.6e %13.6e %13.6e %13.6e",
          prefix, the_stats -> n,        
          the_stats -> mean, the_stats -> sigma_mean, the_stats -> sigma,
                the_stats -> sum_squared);
  
        the_stats = min_stats + i;
        print("%s MinD: %5ld # %13.6e %13.6e %13.6e %13.6e",
            prefix, the_stats -> n, the_stats -> mean,
            the_stats -> sigma_mean, the_stats -> sigma,
                the_stats -> sum_squared);
                
        the_stats = max_stats + i;
        print("%s MaxD: %5ld # %13.6e %13.6e %13.6e %13.6e",
            prefix, the_stats -> n, the_stats -> mean,
            the_stats -> sigma_mean, the_stats -> sigma,
                the_stats -> sum_squared);
                
    }
#if DEBUG_DUMP_STATS_TO_OUTPUT
    print("dump_stats_to_output(): exit.");
#endif
}

